
# $Id: dp-eb-estimates.R,v 1.18 2005/12/08 18:30:32 jon Exp $
# $Source: /projects/starl/cvs/dp-eb/src/dp-eb-estimates.R,v $

###
### Copyright (c) 2004, 2005, David M. Blei and Jon D. McAuliffe.
### This software is distributed under the terms of the GNU
### General Public License.
###

# in the global environment:
#
# B         number of gibbs samples used to construct g0.hat
# N         number of observations
# gsamples  list of length B (except at round 0, where length is 1)
#
# gsamples[[b]] is a
#  list(classes = numeric(N),
#       theta.stars = numeric(N+M),
#       counts = numeric(N+M))
#
# dg0.hat.ideal is a kde constructed using the actual theta.stars in the
#   simulated dataset.


dp.eb.estimates <- function(first = FALSE)
{
  overall.min <-
    min(unlist(lapply(gsamples,
                      function(gs) {
                        min(gs$theta.stars[gs$counts > 0])
                      })));
  overall.max <-
    max(unlist(lapply(gsamples,
                      function(gs) {
                        max(gs$theta.stars[gs$counts > 0])
                      })));
  overall.range <- overall.max - overall.min;

  grid.left  <- overall.min - 0.1 * overall.range;
  grid.right <- overall.max + 0.1 * overall.range;

  kde.results <-
    lapply(gsamples,
           function(gs) {
             locs <- gs$theta.stars[gs$counts > 0];
             d <- density(locs,
                          bw = "SJ-dpi", kernel = "gaussian", n = 1024,
                          from = grid.left, to = grid.right);
             list(locs = locs, d = d);
           });

  x.grid <- kde.results[[1]]$d$x;
  grid.step <- x.grid[2] - x.grid[1];
  y.grid <- rowMeans(sapply(kde.results, function(kde) { kde$d$y }));

  dg0.hat <- function(x)
  {
    res <- numeric(length(x));
    good <- ( (x >= x.grid[1]) & (x < x.grid[length(x.grid)]) );
    res[!good] <- 0;

    if (!any(good))
    {
      return(res);
    }

    left.bdys <- sapply(x[good], function(v) max(which(v >= x.grid)));
    lambdas <- (x[good] - x.grid[left.bdys]) / grid.step;
    res[good] <-
      (1 - lambdas) * y.grid[left.bdys] + lambdas * y.grid[left.bdys + 1];
    res;
  }
  attr(dg0.hat, "x.grid") <- x.grid;
  attr(dg0.hat, "y.grid") <- y.grid;

  if (first)
  {
    rg0.hat <- function(n)
    {
      locs <- kde.results[[1]]$locs[sample(kde.results[[1]]$d$n, n,
                                           replace = TRUE)];
      # the kernel is a gaussian.
      rnorm(n = n, mean = locs, sd = kde.results[[1]]$d$bw);
    }

    if (is.null(ALPHA.INIT))
    {
      # no finite value of alpha will make E[ #clusters | alpha] = # data
      # points, so we use a different strategy at round 0: split the
      # difference between the largest and smallest plausible value of
      # alpha.
      alpha.hat <-
        (kde.results[[1]]$d$n + 1) / (2 * log(kde.results[[1]]$d$n));

    } else {

      alpha.hat <- ALPHA.INIT;
    }
  } else {

    rg0.hat <- function(n)
    {
      kde.idxs <- sample(B, n, replace = TRUE);
      sapply(kde.idxs,
             function(kde.idx) {
               kde <- kde.results[[kde.idx]];
               loc <- kde$locs[sample(kde$d$n, 1)];
               # the kernel is a gaussian.
               rnorm(n = 1, mean = loc, sd = kde$d$bw);
             });
    }

    # this requires the kdes to be constructed first, to avoid duplicating
    # the work of counting the number of theta.stars in each gibbs iterate.
    post.mean.cluster.cnt <-
      mean(sapply(kde.results, function(kder) kder$d$n));

    alpha.hat <-
      uniroot(function(alpha)
              sum(alpha / (alpha + 0:(N-1))) - post.mean.cluster.cnt,
              interval = c(EPS, N))$root;
  }

  return(list(dg0.hat = dg0.hat, rg0.hat = rg0.hat,
              alpha.hat = alpha.hat, kde.results = kde.results));
}


data.sd.m.step <- function(first = FALSE)
{
  if (first)
  {
    return(sd(dataset));
  }
  sqrt(mean(sapply(gsamples,
                   function(gs) {
                     mean((dataset - gs$theta.stars[gs$classes])**2);
                   })));
}


plot.g0.hat <- function(eb.results, ytop, xlim, rescaler = 1, title = "g0",
                        multi = TRUE, show.truth.curves = TRUE, xlab)
{
  if (missing(xlim))
  {
    xlim <- range(attr(eb.results$dg0.hat, "x.grid") / rescaler);
  }

  if (missing(ytop))
  {
    ytop <- max(attr(eb.results$dg0.hat, "y.grid") * rescaler);
    if (multi)
    {
      ytop <-
        max(ytop,
            sapply(eb.results$kde.results,
                   function(kder) max(kder$d$y) * rescaler));
    }
  }

  if (missing(xlab))
  {
    xlab <- "";
  }

  plot(0, 0, type = "n", xlim = xlim, ylim = c(0, ytop), main = title,
       xlab = "");
  mtext(xlab, side = 1, line = 2.3);
#  plot.new();
#  plot.window(xlim = xlim, ylim = c(0, ytop));
#  axis(1); axis(2);
#  title(main = title, xlab = xlab);

  if (multi)
  {
    line.mat <- sapply(eb.results$kde.results, function(v) { v$d$y });
    envelope <- apply(line.mat, 1, range) * rescaler;
    segments(attr(eb.results$dg0.hat, "x.grid") / rescaler, envelope[1,],
             attr(eb.results$dg0.hat, "x.grid") / rescaler, envelope[2,],
             col = "lightgray");

    #for (i in 1:B)
    #{
    #  lines(eb.results$kde.results[[i]]$d$x / rescaler,
    #        eb.results$kde.results[[i]]$d$y * rescaler,
    #        col = "lightgray", lwd = 1);
    #}
  }

  if (show.truth.curves)
  {
    lines(attr(eb.results$dg0.hat, "x.grid") / rescaler,
          dg0(attr(eb.results$dg0.hat, "x.grid")) * rescaler,
          col = "red", lwd = 2);
    lines(attr(eb.results$dg0.hat, "x.grid") / rescaler,
          dg0.hat.ideal(attr(eb.results$dg0.hat, "x.grid")) * rescaler,
          col = "green", lwd = 2);
  }
  lines(attr(eb.results$dg0.hat, "x.grid") / rescaler,
        attr(eb.results$dg0.hat, "y.grid") * rescaler, col = "black", lwd = 2);
}


grayscale.plot.g0.hat <- function(eb.results, ytop, title = "g0", multi = TRUE,
                                  show.truth.curves = TRUE)
{
  plot.new();
  plot.window(xlim = c(min(attr(eb.results$dg0.hat, "x.grid")),
              max(attr(eb.results$dg0.hat, "x.grid"))),
              ylim = c(0, ytop));
  axis(1); axis(2);
  title(main = title);

  if (multi)
  {
    line.mat <- sapply(eb.results$kde.results, function(v) { v$d$y });
    envelope <- apply(line.mat, 1, range);
    segments(attr(eb.results$dg0.hat, "x.grid"), envelope[1,],
             attr(eb.results$dg0.hat, "x.grid"), envelope[2,],
             col = "lightgray");

#    lapply(1:B, function(i) lines(eb.results$kde.results[[i]]$d,
#                                  col = "lightgray",
#                                  lty = "solid", lwd = 1));
  }

  if (show.truth.curves)
  {
    lines(attr(eb.results$dg0.hat, "x.grid"),
          dg0(attr(eb.results$dg0.hat, "x.grid")),
          col = "black", lwd = 2, lty="dashed");
    lines(attr(eb.results$dg0.hat, "x.grid"),
          dg0.hat.ideal(attr(eb.results$dg0.hat, "x.grid")),
          col = "black", lwd = 2, lty="dotted");
  }
  lines(attr(eb.results$dg0.hat, "x.grid"),
        attr(eb.results$dg0.hat, "y.grid"),
        col = "black", lwd = 2);
}
